/* ########################################################################

   PICsim - PIC simulator

   ########################################################################

   Copyright (c) : 2008  Luis Claudio Gamba Lopes

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   For e-mail suggestions :  lcgamboa@yahoo.com
   ######################################################################## */


#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include"picsim.h"
#include"p16fxxx.h"

int
parse_hex(char *line,int bytes)
{
  char snum[200];
  int num;
    
  strncpy(snum,line,bytes);
  snum[bytes]='\0';
  sscanf(snum,"%X",&num);

  return num; 
};


unsigned char
checksum(char* str)
{
  unsigned int i;
  unsigned char acum=0;

  for(i=0; i < ((strlen(str)-1)/2);i++)
  {
    acum+=parse_hex(str+1+(i*2),2);
  }
  
  return acum;
}


int
read_ihx(pic16f * pic,const char * fname, int lrom)
{
  FILE* fin;
  int  lc=0;
  unsigned int bc;
  char line[256];
  unsigned int nbytes,addr,type;  
  
  unsigned short addrh=0;
  unsigned short addrl=0;

  fin=fopen(fname,"r");

  if(fin)
  {
    do
    {
      fgets(line,256,fin);
      lc++;
  
      /*for dos file*/ 
      if(line[strlen(line)-2]=='\r')
      {
        line[strlen(line)-2]='\n';
        line[strlen(line)-1]=0;
      } 

      if(checksum(line) == 0)
      {
        nbytes=parse_hex(line+1,2);
        addr=parse_hex(line+3,4);
        type=parse_hex(line+7,2);

	switch(type)
	{
	  case 0:
	    addrl=addr/2;

            if((((addrh<<16)|addrl)<<1) == 0x400E )
	    {
	      pic->config[0]=(parse_hex(line+9+2,2)<<8)|parse_hex(line+9,2);
	    }
            else
	    {
              
              if((((addrh<<16)|addrl)<<1) >= 0x4200 )
              {
                if(lrom == 1) 
                for(bc=0;bc < nbytes;bc+=2)
	        {
	          pic->eeprom[((addrh<<16)|addrl)+(bc/2)-0x2100]=(parse_hex(line+9+(bc*2)+2,2)<<8)|parse_hex(line+9+(bc*2),2);
                }
              }
              else
              { 
                for(bc=0;bc < nbytes;bc+=2)
	        {
	          pic->prog[((addrh<<16)|addrl)+(bc/2)]=(parse_hex(line+9+(bc*2)+2,2)<<8)|parse_hex(line+9+(bc*2),2);
                }
              }
	    }
	  break;
	  case 1:
            fclose(fin);  
	    return 1;
	  break;
	  case 4:
	    addrh=addr/2;
	  break;
	}
      }
      else
      {
        printf("ERRO: Picsim->File bad checksum line %i!(%s)\n",lc,fname);
        fclose(fin);  
        return 0;
      }
     }
     while(!feof(fin));
     fclose(fin);  
  }
  else
  {
    printf("ERRO: Picsim->File not found!(%s)\n",fname);
    return 0;
  }
  return 1;
};

