/* ########################################################################

   PICsim - PIC simulator

   ########################################################################

   Copyright (c) : 2008  Luis Claudio Gamba Lopes

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   For e-mail suggestions :  lcgamboa@yahoo.com
   ######################################################################## */

#include<stdio.h>
#include<stdlib.h>
#include<string.h>

#include"picsim.h"
#include"p16fxxx.h"
#include"periferic.h"


int read_ihx(pic16f * pic,const char * fname,int lrom);

void
pic_decode_16(pic16f * pic,int print)
{
  unsigned short temp;
  unsigned short opc = pic->prog[pic->pc];
  unsigned short bank= (pic->ram[STATUS] & 0x0060) <<2;
  unsigned char * status = &pic->ram[bank|(STATUS & 0x007F)];
  unsigned char * intcon = &pic->ram[bank|(INTCON & 0x007F)];
  unsigned char afsr=pic->ram[((STATUS & 0x0080)<<1)|FSR];
  unsigned short  pc_ant;
  
  pic->lram=0;
  pic->rram=0;
  
  if(pic->pwd == 1)
  {
    return;
  }
  
  //incompleto - verificar salto pcl e pch!
  if(pic->s2 == 1)
  {
    pic->s2=0;
    return;
  };
  
 
  //print (Address)
  if(print)
  {
    printf("pc=%#06X\t",pic->pc); 
    printf("prog=%#06X\t",pic->prog[pic->pc]); 
  }

  pic->pc++;
  if(pic->pc >= pic->ROMSIZE) pic->pc=0;

//pc
  temp = pic->pc&0x00FF;
  pic->ram[(0x0000)|(PCL & 0x007F)]=temp;
  pic->ram[(0x0080)|(PCL & 0x007F)]=temp;
  pic->ram[(0x0100)|(PCL & 0x007F)]=temp;
  pic->ram[(0x0180)|(PCL & 0x007F)]=temp;
  
  pc_ant = temp;

//indf
  pic->ram[(0x0000)|(INDF & 0x007F)]=pic->ram[afsr];
  pic->ram[(0x0080)|(INDF & 0x007F)]=pic->ram[afsr];
  pic->ram[(0x0100)|(INDF & 0x007F)]=pic->ram[afsr];
  pic->ram[(0x0180)|(INDF & 0x007F)]=pic->ram[afsr];

/*
//0x70 - 0x7F
  for(i=0;i<16;i++)
  {
    temp = pic->ram[bank|(0x70+i)];
    pic->ram[((0x0000)<<2)|(0x70+i)]=temp;
    pic->ram[((0x0020)<<2)|(0x70+i)]=temp;
    pic->ram[((0x0040)<<2)|(0x70+i)]=temp;
    pic->ram[((0x0060)<<2)|(0x70+i)]=temp;
  }

*/
  if(0x0070 != (bank|0x70)) memcpy(&pic->ram[0x0070],&pic->ram[bank|0x70],16);
  if(0x00F0 != (bank|0x70)) memcpy(&pic->ram[0x00F0],&pic->ram[bank|0x70],16);
  if(0x0170 != (bank|0x70)) memcpy(&pic->ram[0x0170],&pic->ram[bank|0x70],16);
  if(0x01F0 != (bank|0x70)) memcpy(&pic->ram[0x01F0],&pic->ram[bank|0x70],16);



  switch(opc & 0x3000)
  {
    case 0x0000:
      switch(opc & 0x0F00)
      {
        case 0x0000:
          switch(opc & 0x0080)
          {
            case 0x0000:
              switch(opc & 0x007F)
              { 
                case 0x0000:
                case 0x0020:
                case 0x0040:
                case 0x0060:
//NOP     --  	No Operation                 1     0000000 xx00000  
                  if(print)printf("NOP\n");
		break;
		case 0x0008:
///RETURN  --  	Return from Subroutine       2     0000000 0001000
                  if(print)printf("RETURN\n");
	          pic->pc=pic->stack[0];
                  for(temp=0;temp<7;temp++)
	            pic->stack[temp]=pic->stack[temp+1];
		  pic->stack[7]=0;
	          pic->s2=1;
                break;
		case 0x0063:
//SLEEP   --  	Go into Standby mode         1     0000000 1100011   	TO,PD
  		  pic->wdt=0;
  		  pic->pwd=1;
	          *status&=~0x08;
	          *status|=0x10;
                break;
		case 0x0064:
//CLRWDT  --  	Clear Watchdog Timer         1     0000000 1100100   	TO,PD
  		  pic->wdt=0;
	          *status|=0x08;
	          *status|=0x10;
  		break;
		case 0x0009:
//RETFIE  --  	Return from interrupt        2     0000000 0001001                                                         
                  if(print)printf("RETFIE\n");
	          pic->pc=pic->stack[0];
                  for(temp=0;temp<7;temp++)
	            pic->stack[temp]=pic->stack[temp+1];
		  pic->stack[7]=0;
	          pic->s2=1;
	          *intcon|=0x80;
                break;
              };
	    break;
	    case 0x0080:
//MOVWF   f  	Move W to f                  1     0000001 fffffff
              if(print)printf("MOVWF %#04X\n",opc & 0x007F);
	      pic->ram[bank|(opc & 0x007F)]=pic->w;
              pic->lram=bank|(opc & 0x007F);
	    break;
	  }
	break;
	case 0x0100:
          switch(opc & 0x0080)
          {
            case 0x0000:
//CLRW    --  	Clear W                      1     0000010 xxxxxxx 	Z	
              if(print)printf("CLRW\n");
              pic->w=0;
            break;
	    case 0x0080:
//CLRF    f  	Clear f                      1     0000011 fffffff 	Z       2
              if(print)printf("CLRF %#04X\n",opc & 0x007F);
              pic->ram[bank|(opc & 0x007F) ]=0;
	    break;
	  };
	break;
	case 0x0200:
//SUBWF   f, d 	Subtract W from f            1     000010 dfffffff 	C,DC,Z  1, 2
          if(print)printf("SUBWF %#04X,%d\n",opc & 0x007F,(opc & 0x0080)>>7);
	
	  if((0xF0&((0x0F&(~pic->w+1))+(0x0F&pic->ram[bank|(opc & 0x007F)]))) !=0)
	    *status|=0x02;
	  else   
	    *status&=~0x02;

	  temp=(~pic->w+1)+pic->ram[bank|(opc & 0x007F)];
	
	  if ((0xFF00&temp) != 0) 
	    *status&=~0x01;
	  else   
	    *status|=0x01;

          if((opc & 0x0080) == 0 )
	  {
	    pic->w = (unsigned char)(0x00FF &temp);
	    if (pic->w == 0) 
	       *status|=0x04;
	    else   
	       *status&=~0x04;
          }
	  else
	  {
	    pic->ram[bank|(opc & 0x007F)]=(unsigned char)(0x00FF &temp);
	    if (pic->ram[bank|(opc & 0x007F) ] == 0) 
	       *status|=0x04;
	    else   
	       *status&=~0x04;
	  };     
        break;
	case 0x0300:
//DECF    f, d 	Decrement f                  1     000011 dfffffff  	Z       1, 2
          if(print)printf("DECF %#04X,%d\n",opc & 0x007F,(opc & 0x0080)>>7);
          if((opc & 0x0080) == 0 )
	  {
	    pic->w=pic->ram[bank|(opc & 0x007F)]-1;
	    if (pic->w == 0) 
	       *status|=0x04;
	    else   
	       *status&=~0x04;
          }
	  else
	  {
	    pic->ram[bank|(opc & 0x007F) ]--;
	    if (pic->ram[bank|(opc & 0x007F) ] == 0) 
	       *status|=0x04;
	    else   
	       *status&=~0x04;
	  };     
        break;
	case 0x0400:
//IORWF   f, d 	Inclusive OR W with f        1     000100 dfffffff   	Z       1, 2
          if(print)printf("IORWF %#04X,%d\n",opc & 0x007F,(opc & 0x0080)>>7);
          if((opc & 0x0080) == 0 )
	  {
	    pic->w= pic->w | pic->ram[bank|(opc & 0x007F)];
	    if (pic->w == 0) 
	       *status|=0x04;
	    else   
	       *status&=~0x04;
          }
	  else
	  {
	    pic->ram[bank|(opc & 0x007F) ]= pic->w | pic->ram[bank|(opc & 0x007F)];
	    if (pic->ram[bank|(opc & 0x007F) ] == 0) 
	       *status|=0x04;
	    else   
	       *status&=~0x04;
	  };     
        break;
	case 0x0500:
//ANDWF   f, d 	AND W with f                 1     000101 dfffffff	Z       1, 2
          if(print)printf("ANDWF %#04X,%d\n",opc & 0x007F,(opc & 0x0080)>>7);
          if((opc & 0x0080) == 0 )
	  {
	    pic->w= pic->w & pic->ram[bank|(opc & 0x007F)];
	    if (pic->w == 0) 
	       *status|=0x04;
	    else   
	       *status&=~0x04;
          }
	  else
	  {
	    pic->ram[bank|(opc & 0x007F) ]= pic->w & pic->ram[bank|(opc & 0x007F)];
	    if (pic->ram[bank|(opc & 0x007F) ] == 0) 
	       *status|=0x04;
	    else   
	       *status&=~0x04;
	  };     
        break;
	case 0x0600:
//XORWF   f, d 	Exclusive OR W with f        1     000110 dfffffff  	Z       1, 2
          if(print)printf("XORWF %#04X,%d\n",opc & 0x007F,(opc & 0x0080)>>7);
          if((opc & 0x0080) == 0 )
	  {
	    pic->w= pic->w ^ pic->ram[bank|(opc & 0x007F)];
	    if (pic->w == 0) 
	       *status|=0x04;
	    else   
	       *status&=~0x04;
          }
	  else
	  {
	    pic->ram[bank|(opc & 0x007F) ]= pic->w ^ pic->ram[bank|(opc & 0x007F)];
	    if (pic->ram[bank|(opc & 0x007F) ] == 0) 
	       *status|=0x04;
	    else   
	       *status&=~0x04;
	  };     
        break;
	case 0x0700:
//ADDWF	f, d 	Add W and f                  1     000111 dfffffff	C,DC,Z  1, 2
          if(print)printf("ADDWF %#04X,%d\n",opc & 0x007F,(opc & 0x0080)>>7);
	
	  if((0xF0&((0x0F&pic->w)+(0x0F&pic->ram[bank|(opc & 0x007F)]))) !=0)
	    *status|=0x02;
	  else   
	    *status&=~0x02;

	  temp=pic->w+pic->ram[bank|(opc & 0x007F)];
	
	  if ((0xFF00&temp) != 0) 
	    *status|=0x01;
	  else   
	    *status&=~0x01;

          if((opc & 0x0080) == 0 )
	  {
	    pic->w = (unsigned char)(0x00FF &temp);
	    if (pic->w == 0) 
	       *status|=0x04;
	    else   
	       *status&=~0x04;
          }
	  else
	  {
	    pic->ram[bank|(opc & 0x007F)]=(unsigned char)(0x00FF &temp);
	    if (pic->ram[bank|(opc & 0x007F) ] == 0) 
	       *status|=0x04;
	    else   
	       *status&=~0x04;
	  };     
        break;
	case 0x0800:
//MOVF    f, d 	Move f                       1     001000 dfffffff 	Z       1, 2
          if(print)printf("MOVF %#04X,%d\n",opc & 0x007F,(opc & 0x0080)>>7);
          if((opc & 0x0080) == 0 )
	  {
	    pic->w=pic->ram[bank|(opc & 0x007F)];
	    if (pic->w == 0) 
	       *status|=0x04;
	    else   
	       *status&=~0x04;
            pic->rram=bank|(opc & 0x007F);
          }
	  else
	  {
	    if (pic->ram[bank|(opc & 0x007F) ] == 0) 
	       *status|=0x04;
	    else   
	       *status&=~0x04;
	  };     
        break;
	case 0x0900:
//COMF    f, d 	Complement f                 1     001001 dfffffff 	Z       1, 2
          if(print)printf("COMF %#04X,%d\n",opc & 0x007F,(opc & 0x0080)>>7);
          if((opc & 0x0080) == 0 )
	  {
	    pic->w=~pic->ram[bank|(opc & 0x007F)];
	    if (pic->w == 0) 
	       *status|=0x04;
	    else   
	       *status&=~0x04;
          }
	  else
	  {
	    pic->ram[bank|(opc & 0x007F) ]=~pic->ram[bank|(opc & 0x007F) ];
	    if (pic->ram[bank|(opc & 0x007F) ] == 0) 
	       *status|=0x04;
	    else   
	       *status&=~0x04;
	  };     
        break;
	case 0x0A00:
//INCF    f, d 	Increment f                  1     001010 dfffffff	Z       1, 2
          if(print)printf("INCF %#04X,%d\n",opc & 0x007F,(opc & 0x0080)>>7);
          if((opc & 0x0080) == 0 )
	  {
	    pic->w=pic->ram[bank|(opc & 0x007F)]+1;
	    if (pic->w == 0) 
	       *status|=0x04;
	    else   
	       *status&=~0x04;
          }
	  else
	  {
	    pic->ram[bank|(opc & 0x007F) ]++;
	    if (pic->ram[bank|(opc & 0x007F) ] == 0) 
	       *status|=0x04;
	    else   
	       *status&=~0x04;
	  };     
        break;
	case 0x0B00:
//DECFSZ  f, d 	Decrement f, Skip if 0       1(2)  001011 dfffffff 		1, 2, 3  
          if(print)printf("DECFSZ %#04X,%d\n",opc & 0x007F,(opc & 0x0080)>>7);
          if((opc & 0x0080) == 0 )
	  {
	    pic->w=pic->ram[bank|(opc & 0x007F)]-1;
	    if (pic->w == 0)
	    {
	      pic->pc++;
	      pic->s2=1;
	    };
          }
	  else
	  {
	    pic->ram[bank|(opc & 0x007F) ]--;
	    if (pic->ram[bank|(opc & 0x007F) ] == 0) 
	    {
	      pic->pc++;
	      pic->s2=1;
	    };
	  };     
        break;
	case 0x0C00:
//RRF     f, d 	Rotate Right f through Carry 1     001100 dfffffff	C       1, 2
          if(print)printf("RRF %#04X,%d\n",opc & 0x007F,(opc & 0x0080)>>7);
	  temp=(pic->ram[bank|(opc & 0x007F)]);
	  if((*status&0x01) ==0x01)temp|=0x0100;
	  if((temp&0x0001) ==0x01)
	    *status|=0x01;
	  else  
	    *status&=~0x01;
	  temp=temp>>1;
          if((opc & 0x0080) == 0 )
	  {
	    pic->w=0x00FF&temp;
          }
	  else
	  {
	    pic->ram[bank|(opc & 0x007F) ]=0x00FF&temp;
	  };     
        break;
	case 0x0D00:
//RLF     f, d 	Rotate Left f through Carry  1     001101 dfffffff	C       1, 2                   
          if(print)printf("RLF %#04X,%d\n",opc & 0x007F,(opc & 0x0080)>>7);
	  temp=(pic->ram[bank|(opc & 0x007F)]);
	  temp=temp<<1;
	  if((*status&0x01) ==0x01)temp|=0x0001;
	  if((temp&0x0100) ==0x0100)
	    *status|=0x0001;
	  else  
	    *status&=~0x0001;
          if((opc & 0x0080) == 0 )
	  {
	    pic->w=0x00FF & temp;
          }
	  else
	  {
	    pic->ram[bank|(opc & 0x007F) ]= 0x00FF & temp;
	  };     
        break;
	case 0x0E00:
//SWAPF   f, d 	Swap nibbles in f            1     001110 dfffffff 		1, 2
          if(print)printf("SWAPF %#04X,%d\n",opc & 0x007F,(opc & 0x0080)>>7);
          if((opc & 0x0080) == 0 )
	  {
	    pic->w=((pic->ram[bank|(opc & 0x007F)]&0x0F)<<4)|
	           ((pic->ram[bank|(opc & 0x007F)]&0xF0)>>4);

          }
	  else
	  {
	    pic->ram[bank|(opc & 0x007F) ]=
	           ((pic->ram[bank|(opc & 0x007F)]&0x0F)<<4)|
	           ((pic->ram[bank|(opc & 0x007F)]&0xF0)>>4);
	  };     
        break;
	case 0x0F00:
//INCFSZ  f, d 	Increment f, Skip if 0       1(2)  001111 dfffffff		1, 2, 3 
          if(print)printf("INCFSZ %#04X,%d\n",opc & 0x007F,(opc & 0x0080)>>7);
          if((opc & 0x0080) == 0 )
	  {
	    pic->w=pic->ram[bank|(opc & 0x007F)]+1;
	    if (pic->w == 0)
	    {
	      pic->pc++;
	      pic->s2=1;
	    };
          }
	  else
	  {
	    pic->ram[bank|(opc & 0x007F) ]++;
	    if (pic->ram[bank|(opc & 0x007F) ] == 0) 
	    {
	      pic->pc++;
	      pic->s2=1;
	    };
	  };     
        break;
      }
    break; 
    case 0x1000:
      switch(opc & 0x0C00)
      {
        case 0x0000:
//BCF     f, b 	Bit Clear f                  1     0100 bbbfffffff           1, 2
          if(print)printf("BCF %#04X,%d\n",opc & 0x007F,(opc & 0x0380)>>7);
	  pic->ram[bank|(opc & 0x007F)]&=~(0x01<<((opc & 0x0380)>>7));
          pic->lram=bank|(opc & 0x007F);
	break;
	case 0x0400:
//BSF     f, b 	Bit Set f                    1     0101 bbbfffffff           1, 2
          if(print)printf("BSF %#04X,%d\n",opc & 0x007F,(opc & 0x0380)>>7);
	  pic->ram[bank|(opc & 0x007F)]|=(0x01<<((opc & 0x0380)>>7));
          pic->lram=bank|(opc & 0x007F);
        break;
	case 0x0800:
//BTFSC   f, b 	Bit Test f, Skip if Clear    1(2)  0110 bbbfffffff           3
        if(print)printf("BTFSC %#04X,%d\n",opc & 0x007F,(opc & 0x0380)>>7);
	if((pic->ram[bank|(opc & 0x007F)] & (0x01<<((opc & 0x0380)>>7))) == 0)
	{
	      pic->pc++;
	      pic->s2=1;
	}
        break;
	case 0x0C00:
//BTFSS   f, b 	Bit Test f, Skip if Set      1(2)  0111 bbbfffffff           3
        if(print)printf("BTFSC %#04X,%d\n",opc & 0x007F,(opc & 0x0380)>>7);
	if((pic->ram[bank|(opc & 0x007F)] & (0x01<<((opc & 0x0380)>>7))) != 0)
	{
	      pic->pc++;
	      pic->s2=1;
	}
        break;
      };
    break; 
    case 0x2000:
      switch(opc & 0x0800)
      {
        case 0x0000:
//CALL    k  	Call subroutine              2     100 kkkkkkkkkkk
          if(print)printf("CALL %#05X\n",opc & 0x07FF);
          for(temp=7;temp>0;temp--)
	    pic->stack[temp]=pic->stack[temp-1];
	  pic->stack[0]=pic->pc;
	  pic->pc=((pic->ram[bank|(PCLATH & 0x007F)]&0x18)<<8)|(opc & 0x07FF);
	  pic->s2=1;
        break;
	case 0x0800:
//GOTO    k  	Go to address                2     101 kkkkkkkkkkk
          if(print)printf("GOTO %#05X\n",opc & 0x07FF);
	  pic->pc=((pic->ram[bank|(PCLATH & 0x007F)]&0x18)<<8)|(opc & 0x07FF);
	  pic->s2=1;
  	break;
      }
    break; 
    case 0x3000:
      switch(opc & 0x0F00)
      {
        case 0x0000:
        case 0x0100:
        case 0x0200:
        case 0x0300:
//MOVLW   k  	Move literal to W            1     1100xx kkkkkkkk 
        if(print)printf("MOVLW %#04X\n",opc & 0x00FF);
	pic->w= (opc & 0x00FF);
        break;
	case 0x0400:
	case 0x0500:
	case 0x0600:
	case 0x0700:
//RETLW   k  	Return with literal in W     2     1101xx kkkkkkkk
          if(print)printf("RETLW %#04X\n",opc & 0x00FF);
	  pic->pc=pic->stack[0];
          for(temp=0;temp<7;temp++)
	    pic->stack[temp]=pic->stack[temp+1];
	  pic->stack[7]=0;
	  pic->s2=1;
	  pic->w= (opc & 0x00FF);
        break;
	case 0x0800:
//IORLW   k  	Inclusive OR literal with W  1     111000 kkkkkkkk   	Z
        if(print)printf("IORLW %#04X\n",opc & 0x00FF);
	pic->w |= (opc & 0x00FF);
	if (pic->w == 0) 
	  *status|=0x04;
	else   
	  *status&=~0x04;
        break;
	case 0x0900:
//ANDLW   k 	AND literal with W           1     111001 kkkkkkkk   	Z
        if(print)printf("ANDLW %#04X\n",opc & 0x00FF);
	pic->w &= (opc & 0x00FF);
	if (pic->w == 0) 
	  *status|=0x04;
	else   
	  *status&=~0x04;
        break;
	case 0x0A00:
//XORLW   k  	Exclusive OR literal with W  1     111010 kkkkkkkk   	Z
        if(print)printf("XORLW %#04X\n",opc & 0x00FF);
	pic->w ^= (opc & 0x00FF);
	if (pic->w == 0) 
	  *status|=0x04;
	else   
	  *status&=~0x04;
        break;
	case 0x0C00:
	case 0x0D00:
//SUBLW   k  	Subtract W from literal      1     11110x kkkkkkkk   	C,DC,Z
        if(print)printf("SUBLW %#04X\n",opc & 0x00FF);
	
	if((0xF0&((0x0F&(~pic->w+1))+(opc & 0x000F))) !=0)
	  *status|=0x02;
	else   
	  *status&=~0x02;

	temp=(~pic->w+1)+(opc & 0x00FF);
	pic->w = (unsigned char)(0x00FF &temp);
	
	if ((0xFF00&temp) != 0) 
	  *status&=~0x01;
	else   
	  *status|=0x01;
	
	if (pic->w == 0) 
	  *status|=0x04;
	else   
	  *status&=~0x04;
        break;
	case 0x0E00:
	case 0x0F00:
//ADDLW   k  	Add literal and W            1     11111x kkkkkkkk   	C,DC,Z
        if(print)printf("ADDLW %#04X\n",opc & 0x00FF);
	
	if((0xF0&((0x0F&pic->w)+(opc & 0x000F))) !=0)
	  *status|=0x02;
	else   
	  *status&=~0x02;

	temp=pic->w+(opc & 0x00FF);
	pic->w = (unsigned char)(0x00FF &temp);
	
	if ((0xFF00&temp) != 0) 
	  *status|=0x01;
	else   
	  *status&=~0x01;
	
	if (pic->w == 0) 
	  *status|=0x04;
	else   
	  *status&=~0x04;
        break;
      }
    break;
  }
  
  /*faz espelhamento de memria*/

  
  temp = pic->ram[bank|(STATUS & 0x007F)];

  pic->ram[(0x0000)|(STATUS & 0x007F)]=temp;
  pic->ram[(0x0080)|(STATUS & 0x007F)]=temp;
  pic->ram[(0x0100)|(STATUS & 0x007F)]=temp;
  pic->ram[(0x0180)|(STATUS & 0x007F)]=temp;
  
  temp = pic->ram[bank|(INTCON & 0x007F)];
  pic->ram[(0x0000)|(INTCON & 0x007F)]=temp;
  pic->ram[(0x0080)|(INTCON & 0x007F)]=temp;
  pic->ram[(0x0100)|(INTCON & 0x007F)]=temp;
  pic->ram[(0x0180)|(INTCON & 0x007F)]=temp;
  
  temp = pic->ram[bank|(PCLATH & 0x007F)];
  pic->ram[(0x0000)|(PCLATH & 0x007F)]=temp;
  pic->ram[(0x0080)|(PCLATH & 0x007F)]=temp;
  pic->ram[(0x0100)|(PCLATH & 0x007F)]=temp;
  pic->ram[(0x0180)|(PCLATH & 0x007F)]=temp;
 

  pic->ram[afsr]=pic->ram[bank|(INDF & 0x007F)];
  
  
/*
//0x70 - 0x7F
  for(i=0;i<16;i++)
  {
    temp = pic->ram[bank|(0x70+i)];
    pic->ram[((0x0000)<<2)|(0x70+i)]=temp;
    pic->ram[((0x0020)<<2)|(0x70+i)]=temp;
    pic->ram[((0x0040)<<2)|(0x70+i)]=temp;
    pic->ram[((0x0060)<<2)|(0x70+i)]=temp;
  //  printf("%#04X ",temp);
  }
  */
  if(0x0070 != (bank|0x70)) memcpy(&pic->ram[0x0070],&pic->ram[bank|0x70],16);
  if(0x00F0 != (bank|0x70)) memcpy(&pic->ram[0x00F0],&pic->ram[bank|0x70],16);
  if(0x0170 != (bank|0x70)) memcpy(&pic->ram[0x0170],&pic->ram[bank|0x70],16);
  if(0x01F0 != (bank|0x70)) memcpy(&pic->ram[0x01F0],&pic->ram[bank|0x70],16);
  //printf("\n"); 
  
  if((pic->ram[bank|(PCL & 0x007F)]) != (pc_ant&0x00FF))
  {
    pic->pc=((pic->ram[bank|(PCLATH & 0x007F)]&0x1F)<<8)|pic->ram[bank|(PCL& 0x007F)];
  }

}


int 
pic_init(pic16f * pic,int processor, const char * fname, int lrom, float freq)
{

   pic->freq=freq;
   pic->processor=processor;
  
   switch(processor)
   {
     case P16F628A:
       pic->ROMSIZE=2048;
       pic->EEPROMSIZE=128;
       pic->RAMSIZE=512;
       pic->PINCOUNT=18;
       pic->IDSIZE=8;
       pic->CONFIGSIZE=1;
       pic->CCPCOUNT=1;
       pic->ADCCOUNT=0;
     break;  
     case P16F877A:
       pic->ROMSIZE=8192;
       pic->EEPROMSIZE=256;
       pic->RAMSIZE=512;
       pic->PINCOUNT=40;
       pic->IDSIZE=8;
       pic->CONFIGSIZE=1;
       pic->CCPCOUNT=2;
       pic->ADCCOUNT=8;
     break;
     default:
       printf("unknow processor!\n");
       return 0;
   }

   pic->ram=calloc(pic->RAMSIZE,sizeof(char));   
   pic->prog=calloc(pic->ROMSIZE,sizeof(short));   
   pic->id=calloc(pic->IDSIZE,sizeof(short));   
   pic->config=calloc(pic->CONFIGSIZE,sizeof(short));   
   pic->eeprom=calloc(pic->EEPROMSIZE,sizeof(char));   
   pic->pins=calloc(pic->PINCOUNT,sizeof(picpin));   
   pic->ccp=calloc(pic->CCPCOUNT,sizeof(char));
   pic->adc=calloc(pic->ADCCOUNT,sizeof(char));
   pic->usart=calloc(2,sizeof(char));

   /*zerar memria*/
   memset(pic->prog,0,pic->ROMSIZE*2); 
   
   if(lrom == 1)
     memset(pic->eeprom,0,pic->EEPROMSIZE); 
   
   pic->pwd=0; 
   
   pic_reset(pic,1);
  
   if(fname == NULL)
     return 0;
   else
     return read_ihx(pic,fname,lrom);
}

int 
pic_reset(pic16f * pic, int flags)
{
  int i;

   //verify MCLRE
   if((flags < 0)&&((pic->config[0] & 0x0020) == 0x0000) )  
   {
     return 0;
   }

   pic->w=0;
   pic->pc=0;
   pic->s2=0;
   pic->lram=0;
   pic->rram=0;

   
   /*zerar memria*/
   memset(pic->ram,0,pic->RAMSIZE); 
   memset(pic->stack,0,8*2); 
   memset(pic->pins,0,sizeof(picpin)*pic->PINCOUNT); 
    
   switch(pic->processor)
   {
     case P16F628A:
       pic->pins[ 0].port=PORTA;pic->pins[ 0].pord=2;
       pic->pins[ 1].port=PORTA;pic->pins[ 1].pord=3;
       pic->pins[ 2].port=PORTA;pic->pins[ 2].pord=4;
       pic->pins[ 3].port=PORTA;pic->pins[ 3].pord=5;
       pic->pins[ 4].port=0    ;pic->pins[ 4].pord=-1;
       pic->pins[ 5].port=PORTB;pic->pins[ 5].pord=0;
       pic->pins[ 6].port=PORTB;pic->pins[ 6].pord=1;
       pic->pins[ 7].port=PORTB;pic->pins[ 7].pord=2;
       pic->pins[ 8].port=PORTB;pic->pins[ 8].pord=3;
       pic->pins[ 9].port=PORTB;pic->pins[ 9].pord=4;
       pic->pins[10].port=PORTB;pic->pins[10].pord=5;
       pic->pins[11].port=PORTB;pic->pins[11].pord=6;
       pic->pins[12].port=PORTB;pic->pins[12].pord=7;
       pic->pins[13].port=0    ;pic->pins[13].pord=-1;
       pic->pins[14].port=PORTA;pic->pins[14].pord=6;
       pic->pins[15].port=PORTA;pic->pins[15].pord=7;
       pic->pins[16].port=PORTA;pic->pins[16].pord=0;
       pic->pins[17].port=PORTA;pic->pins[17].pord=1;

       pic->ccp[0]= 9;
      
       pic->usart[0]=7;
       pic->usart[1]=8;
       break;
     case P16F877A:
       pic->pins[ 0].port=0   ;pic->pins[ 0].pord=-1;
       pic->pins[ 1].port=PORTA;pic->pins[ 1].pord=0;
       pic->pins[ 2].port=PORTA;pic->pins[ 2].pord=1;
       pic->pins[ 3].port=PORTA;pic->pins[ 3].pord=2;
       pic->pins[ 4].port=PORTA;pic->pins[ 4].pord=3;
       pic->pins[ 5].port=PORTA;pic->pins[ 5].pord=4;
       pic->pins[ 6].port=PORTA;pic->pins[ 6].pord=5;
       pic->pins[ 7].port=PORTE;pic->pins[ 7].pord=0;
       pic->pins[ 8].port=PORTE;pic->pins[ 8].pord=1;
       pic->pins[ 9].port=PORTE;pic->pins[ 9].pord=2;
       pic->pins[10].port=0    ;pic->pins[10].pord=-1;
       pic->pins[11].port=0    ;pic->pins[11].pord=-1;
       pic->pins[12].port=0    ;pic->pins[12].pord=-1;
       pic->pins[13].port=0    ;pic->pins[13].pord=-1;
       pic->pins[14].port=PORTC;pic->pins[14].pord=0;
       pic->pins[15].port=PORTC;pic->pins[15].pord=1;
       pic->pins[16].port=PORTC;pic->pins[16].pord=2;
       pic->pins[17].port=PORTC;pic->pins[17].pord=3;
       pic->pins[18].port=PORTD;pic->pins[18].pord=0;
       pic->pins[19].port=PORTD;pic->pins[19].pord=1;
       pic->pins[20].port=PORTD;pic->pins[20].pord=2;
       pic->pins[21].port=PORTD;pic->pins[21].pord=3;
       pic->pins[22].port=PORTC;pic->pins[22].pord=4;
       pic->pins[23].port=PORTC;pic->pins[23].pord=5;
       pic->pins[24].port=PORTC;pic->pins[24].pord=6;
       pic->pins[25].port=PORTC;pic->pins[25].pord=7;
       pic->pins[26].port=PORTD;pic->pins[26].pord=4;
       pic->pins[27].port=PORTD;pic->pins[27].pord=5;
       pic->pins[28].port=PORTD;pic->pins[28].pord=6;
       pic->pins[29].port=PORTD;pic->pins[29].pord=7;
       pic->pins[30].port=0    ;pic->pins[30].pord=-1;
       pic->pins[31].port=0    ;pic->pins[31].pord=-1;
       pic->pins[32].port=PORTB;pic->pins[32].pord=0;
       pic->pins[33].port=PORTB;pic->pins[33].pord=1;
       pic->pins[34].port=PORTB;pic->pins[34].pord=2;
       pic->pins[35].port=PORTB;pic->pins[35].pord=3;
       pic->pins[36].port=PORTB;pic->pins[36].pord=4;
       pic->pins[37].port=PORTB;pic->pins[37].pord=5;
       pic->pins[38].port=PORTB;pic->pins[38].pord=6;
       pic->pins[39].port=PORTB;pic->pins[39].pord=7;
       
       pic->ccp[0]=  17;
       pic->ccp[1]=  16;

       pic->adc[0]=2;
       pic->adc[1]=3;
       pic->adc[2]=4;
       pic->adc[3]=5;
       pic->adc[4]=7;
       pic->adc[5]=8;
       pic->adc[6]=9;
       pic->adc[7]=10;
       
       pic->usart[0]=26;
       pic->usart[1]=25;
       break;
     default:
       break;
    }
   
   for(i=0; i < pic->PINCOUNT; i++)
   {
     pic->pins[i].ptype=PT_CMOS;
     pic->pins[i].dir=PD_IN;
   }  



   if(abs(flags) == 1)
   {
     pic->ram[(0x0000)|(STATUS & 0x007F)]=0x18;
     pic->ram[(0x0080)|(STATUS & 0x007F)]=0x18;
     pic->ram[(0x0100)|(STATUS & 0x007F)]=0x18;
     pic->ram[(0x0180)|(STATUS & 0x007F)]=0x18;
   }	

   pic->ram[TRISA]=0xFF; 
   pic->ram[TRISB]=0xFF; 
   pic->ram[TRISC]=0xFF; 
   pic->ram[TRISD]=0xFF; 
   pic->ram[TRISE]=0xFF; 
   pic->ram[PR2]=0xFF; 

   return 1;
};

void
pic_step(pic16f * pic,int print)
{
  periferic_step_in(pic,print);
  pic_decode_16(pic,print);
  periferic_step_out(pic,print);
}

void
pic_end(pic16f * pic)
{
  if(pic->ram)free(pic->ram);
  if(pic->prog)free(pic->prog);
  if(pic->id)free(pic->id);
  if(pic->config)free(pic->config);
  if(pic->eeprom)free(pic->eeprom);
  if(pic->pins)free(pic->pins);
  if(pic->ccp)free(pic->ccp);
  if(pic->adc)free(pic->adc);
  if(pic->usart)free(pic->usart);
}

unsigned char 
pic_get_pin(pic16f * pic, unsigned char pin)
{
   if((pin-1) < pic->PINCOUNT)
   {
//     if((pic->pins[(pin-1)].type == P_OC)||(pic->pins[(pin-1)].type == P_OT))
//     {
       return pic->pins[(pin-1)].value;
//     }
//     else
//     {
//       return 0;
//     } 
   }
   else
     return 0;  
};


int 
pic_set_pin(pic16f * pic,unsigned char pin,unsigned char value)
{
   unsigned char val;

   if((pin-1) < pic->PINCOUNT)
   {
     if((pic->pins[(pin-1)].dir == PD_IN)&&(pic->pins[(pin-1)].ptype < 3))
     {
          pic->pins[(pin-1)].value=value;
          
          if(pic->pins[(pin-1)].port > 0)
          {
     	      val=0x01<<(pic->pins[(pin-1)].pord);
              if((pic->pins[(pin-1)].value) == 1)
                pic->ram[pic->pins[(pin-1)].port]|=val;
              else	
                pic->ram[pic->pins[(pin-1)].port]&=~val;
	   }

       return 1; 
     }
     else
     { 
       return 0;
     } 
   } 
   else
     return 0;
};


unsigned char 
pic_get_pin_type(pic16f * pic, unsigned char pin)
{
   if((pin-1) < pic->PINCOUNT)
   {
       return pic->pins[(pin-1)].ptype;
   }
   else
     return 0;  
};

unsigned char 
pic_get_pin_dir(pic16f * pic, unsigned char pin)
{
   if((pin-1) < pic->PINCOUNT)
   {
       return pic->pins[(pin-1)].dir;
   }
   else
     return 0;  
};



