/* ########################################################################

   PICsim - PIC simulator

   ########################################################################

   Copyright (c) : 2008  Luis Claudio Gamba Lopes

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   For e-mail suggestions :  lcgamboa@yahoo.com
   ######################################################################## */


#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#ifdef _WIN_
#include <conio.h>
#include <time.h>
#include <windows.h>
#else
#include <termios.h>
#include <sys/ioctl.h>
#endif

#include "picsim.h"
#include "p16fxxx.h"
#include "periferic.h"
        
#define _POSIX_SOURCE 1 /* POSIX compliant source */
#define FALSE 0
#define TRUE 1
        

char SERIALDEVICE[100];

#ifdef _WIN_
HANDLE hCom;
#define BAUDRATE 9600
#else
int fd;
#define BAUDRATE B9600
#endif


int 
serial_open(const char * port)
{
#ifdef _WIN_
hCom = CreateFile(port, GENERIC_READ | GENERIC_WRITE,
0, // exclusive access
NULL, // no security
OPEN_EXISTING,
0, // no overlapped I/O
NULL); // null template

#else
  fd = open(port, O_RDWR | O_NOCTTY | O_NONBLOCK); 
  if (fd < 0) 
  {
    perror(port); 
    return 1; 
  }
#endif
  return 0;
}

int 
serial_close(void)
{
#ifdef _WIN_
  CloseHandle(hCom);
#else    
  if (fd >=0) 
  {
    close(fd);
    fd=-1;
  }
#endif
  return 0;
}



int
serial_cfg(void)
{
#ifdef _WIN_
  BOOL bPortReady;
  DCB dcb;
  COMMTIMEOUTS CommTimeouts;

bPortReady = GetCommState(hCom, &dcb);
dcb.BaudRate = BAUDRATE;
dcb.ByteSize = 8;
dcb.Parity = NOPARITY;
dcb.StopBits = ONESTOPBIT;
dcb.fAbortOnError = TRUE;

// set XON/XOFF
dcb.fOutX = FALSE; // XON/XOFF off for transmit
dcb.fInX = FALSE; // XON/XOFF off for receive
// set RTSCTS
dcb.fOutxCtsFlow = FALSE; // turn off CTS flow control
dcb.fRtsControl = RTS_CONTROL_HANDSHAKE; //
// set DSRDTR
dcb.fOutxDsrFlow = FALSE; // turn off DSR flow control
//dcb.fDtrControl = DTR_CONTROL_ENABLE; //
dcb.fDtrControl = DTR_CONTROL_DISABLE; //
// dcb.fDtrControl = DTR_CONTROL_HANDSHAKE; //

bPortReady = SetCommState(hCom, &dcb);

// Communication timeouts are optional

bPortReady = GetCommTimeouts (hCom, &CommTimeouts);

CommTimeouts.ReadIntervalTimeout = MAXDWORD;
CommTimeouts.ReadTotalTimeoutConstant = 0;
CommTimeouts.ReadTotalTimeoutMultiplier = 0;
CommTimeouts.WriteTotalTimeoutConstant = 0;
CommTimeouts.WriteTotalTimeoutMultiplier = 0;

bPortReady = SetCommTimeouts (hCom, &CommTimeouts);
	

EscapeCommFunction(hCom,SETRTS );

#else
   struct termios newtio;
   int cmd;   
        
//        tcgetattr(fd,&oldtio); /* save current port settings */
        
        bzero(&newtio, sizeof(newtio));
        newtio.c_cflag = BAUDRATE |CS8 | CLOCAL | CREAD;
        newtio.c_iflag = IGNPAR|IGNBRK;
        newtio.c_oflag = 0;
        
        /* set input mode (non-canonical, no echo,...) */
        newtio.c_lflag = 0;
         
        newtio.c_cc[VTIME]    = 0;   /* inter-character timer unused */
        newtio.c_cc[VMIN]     = 5;   /* blocking read until 5 chars received */
        
        tcflush(fd, TCIFLUSH);
        tcsetattr(fd,TCSANOW,&newtio);
        
	cmd=TIOCM_RTS;
	ioctl(fd, TIOCMBIS ,&cmd);
#endif   
	return 0; 
}


      
unsigned long serial_send(unsigned char c)
{
#ifdef _WIN_
   unsigned long nbytes;
    
   WriteFile(hCom, &c, 1, &nbytes,NULL);
   return nbytes;
#else
  return write (fd,&c,1);   
#endif
}

unsigned long serial_rec(unsigned char * c)
{

#ifdef _WIN_
    unsigned long nbytes;
      
    ReadFile(hCom, c, 1,&nbytes, NULL);
#else
    long nbytes;

     nbytes = read (fd,c,1);   
     if(nbytes<0)nbytes=0;
#endif    
    return nbytes;
}

unsigned long serial_rec_tout(unsigned char * c)
{
 unsigned int tout=0;

#ifdef _WIN_
    unsigned long nbytes;
    do
    { 
      Sleep(1);	
      ReadFile(hCom, c, 1,&nbytes, NULL);
#else
    long nbytes;
    do
    { 
      usleep(100);
      nbytes = read (fd,c,1);   
      if(nbytes<0)nbytes=0;
#endif    
      tout++;
    }while((nbytes == 0 )&&(tout < 1000));
    return nbytes;
}


int serialc=0;

void serial(pic16f * pic,int print)
{
  unsigned char temp;
  static int s_open=0;



  if((pic->ram[RCSTA] & 0x80)==0x80)
  {
    if(s_open == 0) 
    {
      if(SERIALDEVICE[0] == 0)
      { 	
#ifdef _WIN_
        strcpy(SERIALDEVICE,"COM2");
#else
        strcpy(SERIALDEVICE,"/dev/tnt2");
#endif
      }

      if(serial_open(SERIALDEVICE) == 0)
      {
        if(print)printf("Port Open:%s!\n",SERIALDEVICE);
        serial_cfg();
        s_open=1;
        pic->ram[TXSTA]|=0x02; //TRMT=1 empty   
      }
      else
      {
        if(print)printf("#Erro Open Port:%s!\n",SERIALDEVICE);
        s_open=-1;
      }
    }
   pic->pins[pic->usart[0]-1].ptype=PT_USART;
   pic->pins[pic->usart[1]-1].ptype=PT_USART;
  }
  else
  {
    if(s_open ==1)
    {
      serial_close();
      s_open=0;
    }
   pic->pins[pic->usart[0]-1].ptype=PT_CMOS;
   pic->pins[pic->usart[1]-1].ptype=PT_CMOS;
  }
  
  if(s_open == 1)
  {

   if(pic->lram == TXREG)    
    { 
      serial_send(pic->ram[TXREG]);
      pic->ram[TXSTA]|=0x02; //TRMT=1 empty   
    }

   if(pic->rram == RCREG)    
    { 
      //clear RCIF
      pic->ram[PIR1]&=~0x20;  
    }

    serialc++;
    if(serialc > 10)
    {
     serialc=0;
  
    if(serial_rec(&temp) == 1)
    {
      pic->ram[RCREG]=temp;
      
      if((pic->ram[PIE1] & 0x20) == 0x20)
      {
        icontrol|=0x0020;
        if(print)printf("serial interrupt (%#04X)\n",temp);
      }
      else
      {
        pic->ram[PIR1]|=0x20;   
      }
    }

    }

  }


}


